﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using Myriadbits.NTFS;
using System.IO;

namespace Myriadbits.DiskSpaceAnalyzer
{
	public partial class TreeFileList : ListView
	{
		private ListViewItem[] m_vecFiles = null;
		private FileData[] m_vecFileDatas = null;

		private Color m_textColorZeroSize = Color.FromArgb(210, 210, 210);
		private Color m_textColorFileInfo = Color.FromArgb(109, 109, 109);
		private ImageList m_imageFiles = new ImageList();

		private TreeViewMode m_treeViewMode = TreeViewMode.AllFiles;
		private UnitTypes m_currentUnit = UnitTypes.Auto;

		private TreeBuilder m_currentSelectedTreeBuilder = null;

		private ToolStripStatusLabel m_tssLabel = null;

		public int SortColumn { get; set; }
		public bool SortAscending { get; set; }
		public Folder CurrentFolder { get; set; }
		
		/// <summary>
		/// Constructor
		/// </summary>
		public TreeFileList()
		{
			InitializeComponent();

			this.MultiSelect = false;
			this.View = System.Windows.Forms.View.Details;
			this.VirtualMode = true;

			this.SortColumn = 0;
			this.SortAscending = true;
			
			// Set the image list to the correct colors
			this.m_imageFiles.ColorDepth = ColorDepth.Depth32Bit;

			// Set the image list
			this.SmallImageList = m_imageFiles;

			// Link the sorting column
			this.ColumnClick += this.onColumnClicked;
			this.RetrieveVirtualItem += this.onRetrieveVirtualItem;
			this.CacheVirtualItems += this.onCacheVirtualItems;
		}


		/// <summary>
		/// Set the status to show info in
		/// </summary>
		/// <param name="tssLabel"></param>
		public void LinkStatusInfo(ToolStripStatusLabel tssLabel)
		{
			m_tssLabel = tssLabel;
		}


		/// <summary>
		/// The selected folder in the tree has changed...
		/// </summary>
		/// <param name="newFolder"></param>
		public void ChangeSelectedFolder(TreeNode node)
		{
			Folder newFolder = node.Tag as Folder;

			// Store the newly selected folder
			this.CurrentFolder = newFolder;

			// Get the rootnode
			this.m_currentSelectedTreeBuilder = null;
			TreeNode rootNode = node;
			while (rootNode.Parent != null) rootNode = rootNode.Parent;
			if (rootNode != null)
			{
				RootFolder rootFolder = rootNode.Tag as RootFolder;
				if (rootFolder != null)
					m_currentSelectedTreeBuilder = rootFolder.Builder;
			}

			// Refresh
			RefreshFileList();
		}


		/// <summary>
		/// Change the current view mode
		/// </summary>
		/// <param name="newViewMode"></param>
		public void ChangeViewMode(TreeViewMode newViewMode)
		{
			if (newViewMode != this.m_treeViewMode)
			{ 
				this.m_treeViewMode = newViewMode;
				RefreshFileList();
			}
		}


		/// <summary>
		/// The current unit type has changed (bytes to MB or GB, or...)
		/// </summary>
		/// <param name="newUnitType"></param>
		public void ChangeUnit(UnitTypes newUnitType)
		{
			if (newUnitType != m_currentUnit)
			{
				m_currentUnit = newUnitType;
				RefreshFileList();
			}
		}


		/// <summary>
		/// Header click -> sort the list
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void onColumnClicked(object sender, ColumnClickEventArgs e)
		{
			if (this.SortColumn == e.Column)
			{
				this.SortAscending = !this.SortAscending;
			}
			else
			{
				this.SortAscending = true;
				this.SortColumn = e.Column;
			}

			// Refresh/update the list
			RefreshFileList();
		}


		/// <summary>
		/// Refresh the virtual file list
		/// </summary>
		public void RefreshFileList()
		{
			// Set the number of files that should be shown in the virtual list
			if (this.CurrentFolder != null)
			{
				this.VirtualListSize = this.CurrentFolder.ChildCount;
				if (this.CurrentFolder.HasFiles)
					this.VirtualListSize += this.CurrentFolder.FileCount;

				// Reset all images
				this.m_imageFiles.Images.Clear();

				// Create a new list of listview items
				this.m_vecFiles = new ListViewItem[this.VirtualListSize];
				this.m_vecFileDatas = new FileData[this.VirtualListSize];

				// Add all child folders
				int fileDataIndex = 0;
				if (this.CurrentFolder.ChildCount > 0)
					foreach (Folder child in this.CurrentFolder.Children)
						this.m_vecFileDatas[fileDataIndex++] = new FileData(child, this.CurrentFolder);

				// Add all files
				if (this.CurrentFolder.FileCount > 0)
				{
					foreach (KeyValuePair<ulong, Myriadbits.NTFS.File> fd in this.CurrentFolder.Files)
					{
						this.m_vecFileDatas[fileDataIndex++] = new FileData(fd.Value, this.CurrentFolder);
					}
				}




				// Create the list with all filedata
				switch (this.SortColumn)
				{
					case 0: // File names
						if (this.SortAscending)
							this.m_vecFileDatas = this.m_vecFileDatas.OrderBy(a => !a.IsFolder).ThenBy(a => a.FileName).ToArray();
						else
							this.m_vecFileDatas = this.m_vecFileDatas.OrderBy(a => a.IsFolder).ThenByDescending(a => a.FileName).ToArray();
						break;
					case 1: // File sizes
						if (this.SortAscending)
							this.m_vecFileDatas = this.m_vecFileDatas.OrderByDescending(a => a.FileSize).ToArray();
						else
							this.m_vecFileDatas = this.m_vecFileDatas.OrderBy(a => a.FileSize).ToArray();
						break;
					case 2: // Number of files
						if (this.SortAscending)
							this.m_vecFileDatas = this.m_vecFileDatas.OrderByDescending(a => a.FileCount).ToArray();
						else
							this.m_vecFileDatas = this.m_vecFileDatas.OrderBy(a => a.FileCount).ToArray();
						break;
					case 3: // File rankings
						if (this.SortAscending)
							this.m_vecFileDatas = this.m_vecFileDatas.OrderBy(a => a.FileRank).ToArray();
						else
							this.m_vecFileDatas = this.m_vecFileDatas.OrderByDescending(a => a.FileRank).ToArray();
						break;
					default:
						break;
				}
		
				if (this.CurrentFolder.Parser != null)
				{
					DriveInfo info = new DriveInfo(this.CurrentFolder.Parser.RootFolder.Name);
					if (info != null && info.TotalSize != 0)
					{
						double percFree = ((double)info.AvailableFreeSpace) / ((double)info.TotalSize);
						double percFull = 1.0 - percFree;

						UnitTypes hardDriveInGb = UnitTypes.GB;
						string freeSt = hardDriveInGb.GetUnitText((double)info.AvailableFreeSpace);
						string totalSt = hardDriveInGb.GetUnitText((double)info.TotalSize);
						string usedSt = hardDriveInGb.GetUnitText((double)info.TotalSize - info.AvailableFreeSpace);

						string driveInfo = string.Format("Current Drive: {0}, Total size: {1}, Free space: {2}, Used: {3} ({4:F1}%)", this.CurrentFolder.Parser.RootFolder.Name, totalSt, freeSt, usedSt, (percFull * 100.0));
						this.m_tssLabel.Text = driveInfo;
						this.m_tssLabel.Visible = true;
					}
				}
			}
			this.Enabled = (this.VirtualListSize > 0);

			this.Invalidate();
		}



		/// <summary>
		/// Virtual file list (faster performance)
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void onRetrieveVirtualItem(object sender, RetrieveVirtualItemEventArgs e)
		{
			e.Item = CreateFileItem(e.ItemIndex);
		}


		/// <summary>
		/// Cache request from virtual list
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void onCacheVirtualItems(object sender, CacheVirtualItemsEventArgs e)
		{
			if (this.CurrentFolder != null && this.m_vecFiles != null)
			{
				//Now we need to rebuild the cache.
				for (int n = e.StartIndex; n <= e.EndIndex; n++) //indexes are inclusive
				{
					CreateFileItem(n);
				}
			}
		}


		/// <summary>
		/// Create a file item to show in the file in the list
		/// </summary>
		/// <param name="itemIndex"></param>
		/// <returns></returns>
		private ListViewItem CreateFileItem(int itemIndex)
		{
			// If the item does not yet exist in the list, create it				
			ListViewItem lvi = this.m_vecFiles[itemIndex];
			if (lvi == null)
			{
				lvi = new ListViewItem();
				FileData fileData = this.m_vecFileDatas[itemIndex];
				lvi.Tag = fileData;
				this.m_vecFiles[itemIndex] = lvi;

				ListViewItem.ListViewSubItem lvsi = null;

				// Load the icon
				fileData.LoadIcon();

				// Set the tooltip to the full filename
				lvi.ToolTipText = string.Format("{0}\nSize: {1} B\nReferenceNumber: {2}", fileData.FileName, fileData.FileSize, fileData.FileReferenceId); // TODO Add extra info??

				// Add the icon to the imagelist			
				fileData.AddIconToImageList(this.m_imageFiles);

				// Set all items			
				lvi.Text = fileData.FileName;
				lvi.ImageIndex = fileData.GetImageIndex(this.m_imageFiles);
				lvi.UseItemStyleForSubItems = false;

				Color infoColor = m_textColorFileInfo;
				if (fileData.FileSize == 0)
					infoColor = m_textColorZeroSize;

				// Paint the file/folder vague when not selected
				
				if (m_currentSelectedTreeBuilder != null)
				{
					bool makeTransparent = false;
					if (fileData.IsFolder)
						makeTransparent = !m_currentSelectedTreeBuilder.IsChildVisible(fileData.ThisFolder, this.m_treeViewMode);
					else
						makeTransparent = !m_currentSelectedTreeBuilder.IsFileVisible(fileData, this.m_treeViewMode);

					if (makeTransparent)
					{
						lvi.ForeColor = m_textColorZeroSize;
						infoColor = m_textColorZeroSize; // All other text should be vague as well....
					}
				}


				// Size
				lvsi = lvi.SubItems.Add(this.m_currentUnit.GetUnitText(fileData.FileSize));
				lvsi.ForeColor = infoColor;


				// Number of files
				if (fileData.IsFolder)
					lvsi = lvi.SubItems.Add(fileData.FileCount.ToString());
				else
					lvsi = lvi.SubItems.Add("");
				lvsi.ForeColor = infoColor;

				// Rank
				if (fileData.IsFolder || fileData.File.Rank < 0)
					lvsi = lvi.SubItems.Add("");
				else
					lvsi = lvi.SubItems.Add((fileData.File.Rank + 1).ToString());
				lvsi.ForeColor = infoColor;

				// Last modification
				//if (fileData.HasStandardInformation)
				//	lvsi = lvi.SubItems.Add(string.Format("{0} {1}", fileData.LastModified.ToString("d"), fileData.LastModified.ToString("HH:mm")));
				//else
				lvsi = lvi.SubItems.Add("");
				//lvsi.ForeColor = infoColor;

				// Attributes
				//if (fileData.HasStandardInformation)
				//{
				//	string fileFlagString = string.Empty;
				//	uint fileFlags = fileData.StandardInformation.DOSFilePermissions;
				//	if ((fileFlags & 0x0001) != 0) fileFlagString += "R"; // Read-only
				//	if ((fileFlags & 0x0002) != 0) fileFlagString += "H"; // Hidden
				//	if ((fileFlags & 0x0004) != 0) fileFlagString += "S"; // System
				//	if ((fileFlags & 0x0020) != 0) fileFlagString += "A"; // Archive
				//	if ((fileFlags & 0x0040) != 0) fileFlagString += "D"; // Device
				//	if ((fileFlags & 0x0080) != 0) fileFlagString += "N"; // Normal
				//	if ((fileFlags & 0x0100) != 0) fileFlagString += "T"; // Temporary
				//	if ((fileFlags & 0x0200) != 0) fileFlagString += "F"; // SparseFile
				//	if ((fileFlags & 0x0400) != 0) fileFlagString += "P"; // Reparse point
				//	if ((fileFlags & 0x0800) != 0) fileFlagString += "C"; // Compressed
				//	if ((fileFlags & 0x1000) != 0) fileFlagString += "O"; // Offline
				//	if ((fileFlags & 0x2000) != 0) fileFlagString += "I"; // Not Content Indexed
				//	if ((fileFlags & 0x4000) != 0) fileFlagString += "E"; // Encrypted
				//	lvsi = lvi.SubItems.Add(fileFlagString);
				//}
				//else
				lvsi = lvi.SubItems.Add("");
				//lvsi.ForeColor = infoColor;
			}
			return lvi;
		}


		/// <summary>
		/// Get the current selected file
		/// </summary>
		/// <returns></returns>
		public FileData SelectedFileData
		{
			get
			{
				if (this.SelectedIndices.Count > 0)
				{
					int itemIndex = this.SelectedIndices[0];
					if (itemIndex >= 0 && itemIndex < m_vecFileDatas.Count())
					{
						FileData fd = this.m_vecFileDatas[itemIndex];
						return fd;
					}
				}
				return null;
			}
		}


		/// <summary>
		/// Delete a file from the listview
		/// </summary>
		public void DeleteFile()
		{
			// Set the number of files that should be shown in the virtual list
			this.VirtualListSize--;
			this.m_vecFiles = new ListViewItem[this.VirtualListSize];

			this.RefreshFileList();
		}


		/// <summary>
		/// Select a file in this list
		/// </summary>
		/// <param name="file"></param>
		public void SelectFile(Myriadbits.NTFS.File fileToSelect)
		{
			// Find the current selected file in this folder
			for (int n = 0; n < m_vecFileDatas.Length; n++)
			{
				if (!m_vecFileDatas[n].IsFolder && m_vecFileDatas[n].File.ReferenceNumber == fileToSelect.ReferenceNumber)
				{
					this.Items[n].Selected = true;
					this.EnsureVisible(n);
					break;
				}
			}
		}

	}
}
