﻿//
// NTFS - Myriadbits .NET NTFS library. 
// Fast MFT reader.
// Copyright (C) 2014 Myriadbits, Jochem Bakker
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// For more information, contact me at: info@myriadbits.com
//

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.IO;
using System.Threading;


namespace Myriadbits.NTFS
{
	public class NTFS
	{
		public const ulong referenceMask = 0x0000FFFFFFFFFFFF;
		public const ulong indexMask     = 0xFFFF000000000000;
	};


	/// <summary>
	/// Parse a volume
	/// </summary>
	public class NTFSVolume
	{
		// Private members
		private UnmanagedFileLoader fileHandle = null;
		private ulong BytesPerSect = 0;
		private ulong SectperCluster = 0;
		private ulong MFTCluster = 0;
		private NTFSDataRuns dataRuns = null;		

		// Events
		public event LogHandler Logger;


		/// <summary>
		/// Constructor
		/// </summary>
		/// <param name="listLog"></param>
		public NTFSVolume()
		{
		}


		/// <summary>
		/// Properties of the volume
		/// </summary>
		public string Drive { get; set; }


		/// <summary>
		/// The capacity of the drive (total space)
		/// </summary>
		public ulong Capacity { get; set; }


		/// <summary>
		/// The name of this volume
		/// </summary>
		public string Name { get; set; }



		/// <summary>
		/// Background process
		/// </summary>
		/// <param name="threadContext"></param>
	    public static bool IsNTFS(string drive)
		{
			try
			{
				// Use the first char
				string driveName = drive.Substring(0, 1);

				// Read NTFS header
				UnmanagedFileLoader fl = new UnmanagedFileLoader(string.Format("\\\\.\\{0}:", driveName));
				byte[] header = fl.ReadBytes(1024); 
			
				// Is this drive a NTFS drive???
				return (header[3] == 'N' && header[4] == 'T' && header[5] == 'F' && header[6] == 'S');
			}
			catch (Exception)
			{
				return false;
			}
		}


		/// <summary>
		/// Returns true if this volume is open
		/// </summary>
		public bool IsOpen 
		{ 
			get
			{
				return fileHandle != null;
			}
		}


		/// <summary>
		/// Open a drive
		/// </summary>
		/// <param name="fileReference"></param>
		/// <returns></returns>
		public bool Open(string drive)
		{
			// Read NTFS header
			this.Drive = drive;
			fileHandle = new UnmanagedFileLoader(string.Format("\\\\.\\{0}:", drive));
			byte[] header = fileHandle.ReadBytes(1024);

			// Get some generic data
			DriveInfo di = new DriveInfo(this.Drive);
			this.Capacity = (ulong) di.TotalSize;
			this.Name = di.VolumeLabel;
			
			// Is this drive a NTFS drive???
			Stopwatch sw = Stopwatch.StartNew();
			bool isNTFS = (header[3] == 'N' && header[4] == 'T' && header[5] == 'F' && header[6] == 'S');
			if (isNTFS)
			{
				// Read general NTFS data
				DataBuffer ntfsheader = new DataBuffer(header);
				BytesPerSect = ntfsheader.GetUShort(0x0B);
				SectperCluster = ntfsheader.GetByte(0x0D);
				MFTCluster = ntfsheader.GetULong(0x30);
				Log("Drive {0}", this.Drive);
				Log("- SectorSize: {0}, SectorsPerCluster: {1}, MFT: {2}", BytesPerSect, SectperCluster, MFTCluster);

				// Determine the offset of the MFT (Master File Table)
				ulong blockSize = 128 * SectperCluster * BytesPerSect;
				ulong offsetRead = MFTCluster * SectperCluster * BytesPerSect;

				// Read part of the MFT
				fileHandle.Seek(offsetRead);

				// Read a part of the MFT
				byte[] data = new byte[blockSize];
				uint reads = fileHandle.ReadFile(ref data, (int) blockSize);
				DataBuffer mftdata = new DataBuffer(data);
				NTFSFileRecord mftFile = new NTFSFileRecord(mftdata, (int) BytesPerSect);

				// Create the dataruns				
				NTFSAttribute dataAttribute = mftFile.Attributes.Where(a => a.AttributeType == NTFSAttributeType.Data).FirstOrDefault();
				dataRuns = new NTFSDataRuns(dataAttribute);

				Log("- Processtime {0} ms", sw.ElapsedMilliseconds);
				Log("- MFT size: {0} Mb", (int) (dataRuns.TotalSize * SectperCluster * BytesPerSect) / (1024*1024));
				Log("");
				return true;
			}
			else
			{
				Log("Sorry, the selected drive is NOT an NTFS drive");
				fileHandle = null;
			}
			return false;
		}


		/// <summary>
		/// Close this volume
		/// </summary>
		public void Close()
		{
			if (IsOpen)
			{
				fileHandle = null;
				this.dataRuns.Clear();
			}
		}


		/// <summary>
		/// Read a single file record
		/// </summary>
		/// <param name="requestedFileRecordReference"></param>
		/// <returns></returns>
		public NTFSFileRecord ReadFileRecord(ulong requestedFileRecordReference)
		{
			if (IsOpen)
			{
				// Loop through the dataruns
				int datarunCounter = 0;
				ulong datarunStartFileRecord = 0;
				foreach (NTFSDataRun dataRun in dataRuns.DataRuns)
				{
					ulong datarunTotalFileRecords = dataRun.Length * SectperCluster * BytesPerSect / 1024;
					Log("  - Datarun {0}, Length: {1}, Offset: {2}, FileRecords: {3}", datarunCounter, dataRun.Length, dataRun.Position, datarunTotalFileRecords);
					if (requestedFileRecordReference >= datarunStartFileRecord && requestedFileRecordReference < datarunStartFileRecord + datarunTotalFileRecords)
					{
						// Requested file reference is in this datarun

						// Calculate the requested position
						ulong offsetRead = dataRun.Position * SectperCluster * BytesPerSect;
						offsetRead += (requestedFileRecordReference - datarunStartFileRecord) * 1024;

						Log("  - Found requested filerecord in datarun starting at {0}", dataRun.Position);
						Log("  - Reading position {0}", offsetRead);
						
						// Read the filerecords
						byte[] data = new byte[1024];				
						fileHandle.Seek(offsetRead);
						uint numberOfBytes = fileHandle.ReadFile(ref data, (int)1024);

						// Create a buffer around the data
						DataBuffer dataBuffer = new DataBuffer(data);
						NTFSFileRecord fileRecord = new NTFSFileRecord(dataBuffer, (int) BytesPerSect);

						// Return the found file record
						return fileRecord;
					}

					// Next datarun
					datarunStartFileRecord += datarunTotalFileRecords;
					datarunCounter++;
				}
			}
			else
				throw new ApplicationException("The volume is not yet opened, please use 'Open' before reading a file record!");
			return null;
		}


		/// <summary>
		/// Send some logging information
		/// </summary>
		/// <param name="format"></param>
		/// <param name="args"></param>
		private void Log(string format, params object[] args) 
		{
			if (this.Logger != null)
				this.Logger(this, string.Format(format, args));
		}


		/// <summary>
		/// Send some logging information
		/// </summary>
		/// <param name="format"></param>
		/// <param name="args"></param>
		public StringBuilder GetInformation(NTFSReader reader, ulong requestedFileRecordReference)
		{
			NTFSFileRecord filerecord = ReadFileRecord(requestedFileRecordReference);
			return filerecord.GetInformation(this, reader);
		}


	}
}
